import os
from io import StringIO
from sys import platform

import tkinter as Tk
from tkinter import filedialog
import tkinter.ttk as ttk

import configparser

from fasta import fasta
from fasta import fastachar_static

CONFIG = dict(linux = dict(INIFILE = 'fastacharrc',
                           INIPATH = '.config/fastachar'),
              win32 = dict(INIFILE = 'fastachar.ini',
                           INIPATH ='.fastachar'))

class ConfigFastachar(object):
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.load()
        
    def get_home(self):
        if platform == 'linux':
            home_dir = os.getenv('HOME')
        elif platform == 'win32':
            home_dir = os.path.join(os.getenv('HOMEDRIVE'),
                                    os.getenv('HOMEPATH'))
        else:
            raise OSError('Detected an unsupported OS.')
        return home_dir
    
    def get_path(self):
        home_dir = self.get_home()
        path = os.path.join(home_dir,
                            CONFIG[platform]['INIPATH'])
        os.makedirs(path, exist_ok=True)
        return os.path.join(path, 
                            CONFIG[platform]['INIFILE'])

    def set_defaults(self):
        default_cwd = self.get_home()
        if not os.path.exists(default_cwd):
            default_cwd = os.getcwd()
        self.config['DEFAULT'] = dict(working_directory=default_cwd)
        
        
    
    def load(self):
        p = self.get_path()
        cfg_file = self.config.read(p)
        if not cfg_file:
            self.set_defaults()
        else:
            try:
                cwd = self.config['DEFAULT']['working_directory']
            except KeyError:
                self.set_defaults()
            else:
                if not os.path.exists(cwd):
                    self.set_defaults()
                else:
                    self.config['DEFAULT'] = dict(working_directory=cwd)

    def save(self):
        p = self.get_path()
        with open(p, 'w') as fp:
            self.config.write(fp)

    
          
class Case(object):
    LIST_KWDS = "species setA setB".split()
    
    def __init__(self):
        self.clear()
        
    def clear(self):
        self.data = {}
        
    def populate(self, filename,
                 species,
                 setA,
                 setB,
                 operation):
        self.data = dict(filename=filename,
                         species=species,
                         setA=setA,
                         setB=setB,
                         operation=operation)

    def parse_line(self, line):
        kwd, value = line.split("=")
        kwd=kwd.strip()
        value=value.strip()
        if kwd in Case.LIST_KWDS:
            value=[i.strip() for i in value.split(",")]
            value.sort()
        return kwd, value
    
    def load(self, filename):
        if not os.path.exists(filename):
            error = fasta.ERROR_FILE_NOT_FOUND
            arg = filename
            return error, arg
        # file exists, now open it.
        error = fasta.OK
        arg = ''
        with open(filename,'r') as fp:
            while True:
                line = fp.readline()
                if not line:
                    break
                try:
                    kwd, value = self.parse_line(line)
                except ValueError:
                    error = fasta.ERROR_CASE
                    arg = line
                else:
                    self.data[kwd]=value
        return error, arg
    
    def save(self, filename):
        with open(filename,'w') as fp:
            fp.write("filename = {}\n".format(self.data['filename']))
            fp.write("species = {}\n".format(" , ".join(self.data['species'])))
            fp.write("setA = {}\n".format(" , ".join(self.data['setA'])))
            fp.write("setB = {}\n".format(" , ".join(self.data['setB'])))
            fp.write("operation = {}\n".format(self.data['operation']))
        

class Gui():
    def __init__(self):
        self.root = Tk.Tk()
        self.root.wm_title('Fastachar')
        self.config = ConfigFastachar()
        self.cwd = self.getcwd()
        self.S = fasta.SequenceData()
        self.case = Case()
        self.reportxls = fasta.ReportXLS()

    def getcwd(self):
        cwd = self.config.config['DEFAULT']['working_directory']
        return cwd

    def setcwd(self,cwd):
        self.config.config['DEFAULT']['working_directory'] = cwd
      
    def about_window(self):
        toplevel = Tk.Toplevel()
        label1 = Tk.Label(toplevel, text=fastachar_static.ABOUT_TEXT, height=0, width=100,
                          justify=Tk.LEFT)
        label1.pack()
        label2 = Tk.Label(toplevel, text=fastachar_static.DISCLAIMER, height=0, width=100,
                          justify=Tk.LEFT)
        label2.pack()
        bt = Tk.Button(toplevel, text="Close", command=toplevel.destroy)
        bt.pack()
        toplevel.focus_force()

    def error_window(self, err_code, arg = ''):
        toplevel = Tk.Toplevel()
        if arg:
            text = " : ".join([fastachar_static.ERRORS[err_code], arg])
        else:
            text = fastachar_static.ERRORS[err_code]
        label1 = Tk.Label(toplevel, text=text,
                          height=0, width=len(text), padx=10, pady=10,
                          justify=Tk.LEFT)
        label1.pack()
        bt = Tk.Button(toplevel, text="Close", padx=10, pady=10,
                       command=toplevel.destroy)
        bt.pack()
        toplevel.focus_force()


    def help_window(self):
        toplevel = Tk.Toplevel()
        frame = Tk.Frame(toplevel)
        frame.pack(side=Tk.TOP, fill=Tk.BOTH, expand=1)
        sb_report = Tk.Scrollbar(frame, orient=Tk.VERTICAL)
        sb_report.pack(side=Tk.RIGHT, fill=Tk.Y)
        report = Tk.Text(frame, state=Tk.DISABLED,
                         padx=10, pady=10, yscrollcommand=sb_report.set)
        report.pack(side=Tk.RIGHT, fill=Tk.BOTH, expand=1)
        sb_report.config(command=report.yview)
        bt = Tk.Button(toplevel, text="Close", command=toplevel.destroy)
        bt.pack(side=Tk.BOTTOM)
        report.config(state=Tk.NORMAL)
        for line in fastachar_static.HELP_TEXT.split("\n"):
            report.insert(Tk.END, "{}\n".format(line))
        report.config(state=Tk.DISABLED)
        toplevel.focus_force()


    def create_menu(self):
        menubar = Tk.Menu(self.root)
        # create a pulldown menu, and add it to the menu bar
        filemenu = Tk.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open fasta file", command=self.cb_open_fasta_file)
        filemenu.add_command(label="Open case file", command=self.cb_open_case_file)
        filemenu.add_command(label="Save case file", command=self.cb_save_case_file)
        filemenu.add_separator()
        filemenu.add_command(label="Set working directory", command=self.cb_set_working_dir)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.root.quit)
        #
        menubar.add_cascade(label="File", menu=filemenu)
        # create more pulldown menus
        helpmenu = Tk.Menu(menubar, tearoff=0)
        helpmenu.add_command(label="About", command=self.cb_about)
        helpmenu.add_command(label="Help", command=self.cb_help)
        menubar.add_cascade(label="Help",  menu=helpmenu)
        outputmenu = Tk.Menu(menubar, tearoff=0)
        outputmenu.add_command(label="Save report (txt)", command=self.cb_save_report)
        outputmenu.add_command(label="Save report (xls)", command=self.cb_save_report_xls)
        menubar.add_cascade(label="Output", menu=outputmenu)
        # display the menu
        self.root.config(menu=menubar)

    def create_layout(self):
        root = self.root
        for i in range(3):
            Tk.Grid.columnconfigure(root, i, weight=1)
        Tk.Grid.rowconfigure(root, 1, weight=1)
        Tk.Grid.rowconfigure(root, 4, weight=1)
            
        cnf = dict(ipadx=10, ipady=10, padx=10, pady=0)
        cnfsticky = dict(sticky=Tk.N+Tk.E+Tk.S+Tk.W)
        Tk.Label(root, text="Species").grid(row=0, column=0, **cnf)
        Tk.Label(root, text="Set A").grid(row=0, column=1, **cnf)
        Tk.Label(root, text="Set B").grid(row=0, column=2, **cnf)
        # the list boxes
        #     define scrollbars
        frame_sequences=Tk.Frame()
        frame_A=Tk.Frame()
        frame_B=Tk.Frame()
        frame_sequences.grid(row=1, column=0, **cnf, **cnfsticky)
        frame_A.grid(row=1, column=1, **cnf, **cnfsticky)
        frame_B.grid(row=1, column=2, **cnf, **cnfsticky)

        frame_sequences.grid_rowconfigure(0,weight=1)
        frame_sequences.grid_columnconfigure(0,weight=1)
        frame_A.grid_rowconfigure(0,weight=1)
        frame_A.grid_columnconfigure(0,weight=1)
        frame_B.grid_rowconfigure(0,weight=1)
        frame_B.grid_columnconfigure(0,weight=1)
        
        # configure the frames with grid, so that the scrollbars can be placed easily.

        # the scrollbars:    
        sb_sequences = Tk.Scrollbar(frame_sequences, orient=Tk.VERTICAL)
        sb_A = Tk.Scrollbar(frame_A, orient=Tk.VERTICAL)
        sb_B = Tk.Scrollbar(frame_B, orient=Tk.VERTICAL)
        sb_hor_sequences = Tk.Scrollbar(frame_sequences, orient=Tk.HORIZONTAL)
        sb_hor_A = Tk.Scrollbar(frame_A, orient=Tk.HORIZONTAL)
        sb_hor_B = Tk.Scrollbar(frame_B, orient=Tk.HORIZONTAL)

        sb_sequences.grid(column=1, row=0, **cnfsticky)
        sb_A.grid(column=1, row=0, **cnfsticky)
        sb_B.grid(column=1, row=0, **cnfsticky)
        sb_hor_sequences.grid(row=1, column=0, **cnfsticky)
        sb_hor_A.grid(row=1, column=0, **cnfsticky)
        sb_hor_B.grid(row=1, column=0, **cnfsticky)
        #     define listboxes with scrollbar attached
        cnf_lb=dict(selectmode=Tk.EXTENDED)
        self.lb_sequences=Tk.Listbox(frame_sequences,
                                     xscrollcommand=sb_hor_sequences.set,
                                     yscrollcommand=sb_sequences.set,
                                     **cnf_lb)
        self.lb_A=Tk.Listbox(frame_A,
                             xscrollcommand=sb_hor_A.set,
                             yscrollcommand=sb_A.set,
                             **cnf_lb)
        self.lb_B=Tk.Listbox(frame_B,
                             xscrollcommand=sb_hor_B.set,
                             yscrollcommand=sb_B.set,
                             **cnf_lb)
        #     connect the xview and yview to the scrollbars
        sb_sequences.config(command=self.lb_sequences.yview)
        sb_A.config(command=self.lb_A.yview)
        sb_B.config(command=self.lb_B.yview)
        sb_hor_sequences.config(command=self.lb_sequences.xview)
        sb_hor_A.config(command=self.lb_A.xview)
        sb_hor_B.config(command=self.lb_B.xview)
        #     place the listboxes
        self.lb_sequences.grid(column=0, row=0, **cnfsticky)
        self.lb_A.grid(column=0, row=0, **cnfsticky)
        self.lb_B.grid(column=0, row=0, **cnfsticky)
        self.data=dict()
        self.data[self.lb_sequences]=[]
        self.data[self.lb_A]=[]
        self.data[self.lb_B]=[]
        
        # operation label and operation radio buttons
        Tk.Label(root, text="Operation").grid(row=2, column=0, **cnf)
        frame = Tk.Frame(root)
        frame.grid(row=3, column=0, **cnf)
        self.operation_method = Tk.IntVar()
        self.operation_method.set(1)
        Tk.Radiobutton(frame, text="Unique characters A",
                       variable=self.operation_method, value=1).pack(anchor=Tk.W)
        Tk.Radiobutton(frame, text="Unique characters B",
                       variable=self.operation_method, value=2).pack(anchor=Tk.W)
        Tk.Radiobutton(frame, text="Differences within A",
                       variable=self.operation_method, value=3).pack(anchor=Tk.W)
        Tk.Radiobutton(frame, text="Differences within B",
                       variable=self.operation_method, value=4).pack(anchor=Tk.W)

        bt_run = Tk.Button(root, text="Process", command=self.cb_run)
        bt_run.grid(row=3, column=1, **cnf)

        bt_clr = Tk.Button(root, text="Clear output", command=self.cb_clr)
        bt_clr.grid(row=3, column=2, **cnf)

        frame_report = Tk.Frame(pady=10)
        frame_report.grid(row=4, column=0, columnspan=3,
                          sticky=Tk.W+Tk.E+Tk.N+Tk.S, padx=10,pady=10)
        frame_report.grid_rowconfigure(0, weight=1)
        frame_report.grid_columnconfigure(0, weight=1)
        
        sb_report = Tk.Scrollbar(frame_report, orient=Tk.VERTICAL)
        sb_hor_report = Tk.Scrollbar(frame_report, orient=Tk.HORIZONTAL)
        sb_report.grid(column=1, row=0, **cnfsticky)
        sb_hor_report.grid(column=0, row=1, **cnfsticky)
        self.report = Tk.Text(frame_report, state=Tk.DISABLED,
                              wrap=None,
                              yscrollcommand=sb_report.set,
                              xscrollcommand=sb_hor_report.set)
        self.report.grid(column=0, row=0, **cnfsticky)
        sb_report.config(command=self.report.yview)
        sb_hor_report.config(command=self.report.xview)

    def create_bindings(self):
        self.dragging=False
        self.lb_sequences.bind("<B3-Motion>", self.cb_b1_motion_lb)
        self.lb_sequences.bind("<ButtonRelease-3>", self.cb_b1_release_lb)
        self.lb_A.bind("<B3-Motion>", self.cb_b1_motion_lb)
        self.lb_A.bind("<ButtonRelease-3>", self.cb_b1_release_lb)
        self.lb_B.bind("<B3-Motion>", self.cb_b1_motion_lb)
        self.lb_B.bind("<ButtonRelease-3>", self.cb_b1_release_lb)


    def cb_b1_motion_lb(self, event):
        self.dragging=True
        
    def cb_b1_release_lb(self, event):
        if self.dragging:
            s = self.release_in_listbox(event)
            if s and s!=event.widget:
                self.move_items(event.widget, s)
        self.dragging=False
        
        
    def move_items(self, lb_from, lb_to):
        items = list(lb_from.curselection())
        items.reverse()
        for i in items:
            lb_from.delete(i)
            s=self.data[lb_from].pop(i)
            self.data[lb_to].append(s)
            self.data[lb_to].sort()
            j=self.data[lb_to].index(s)
            lb_to.insert(j,s)
            
        
    def release_in_listbox(self, event):
        lbs=[self.lb_sequences, self.lb_A, self.lb_B]
        x = event.x_root
        y = event.y_root
        s=0
        for lb in lbs:
            x0 = lb.winfo_rootx()
            y0 = lb.winfo_rooty()
            dx = lb.winfo_width()
            dy = lb.winfo_height()
            if x>=x0 and x<=x0+dx and y>=y0 and y<=y0+dy:
                s = lb
        return s
            
    
        
    def populate_list_with_items(self, items, lb, delete_all=True):
        if delete_all:
            lb.delete(0,Tk.END)
        for item in items:
            lb.insert(Tk.END, item)
        
    def cb_open_fasta_file(self):
        self.fasta_file = filedialog.askopenfilename(defaultextension=".fas",
                                                     filetypes=[('fasta files', '.fas'), ('all files', '.*')],
                                                     initialdir=self.cwd,
                                                     #initialfile,
                                                     multiple=False,
                                                     #message,
                                                     parent=self.root,
                                                     title="Open fasta file")
        r,arg = self.open_fasta_file()
        if r != fasta.OK:
            if arg:
                arg += " (%s)"%(self.fasta_file)
            else:
                arg = self.fasta_file
            self.error_window(r, arg=arg)
            
    def cb_open_case_file(self):
        case_file = filedialog.askopenfilename(defaultextension=".fc",
                                               filetypes=[('case files', '.fc'), ('all files', '.*')],
                                               initialdir=self.cwd,
                                               #initialfile,
                                               multiple=False,
                                               #message,
                                               parent=self.root,
                                               title="Open case file")
        if case_file:
            r, arg = self.open_case_file(case_file)
            if r != fasta.OK:
                self.error_window(r, arg=case_file)
        # else ignore silently

        
    def open_fasta_file(self):
        if self.fasta_file:
            error, arg = self.S.load(self.fasta_file)
            if error == fasta.OK:
                try:
                    species = self.S.get_species_list()
                except:
                    error = fasta.ERROR_FILE_INVALID
                    arg = ''
                else:
                    species.sort()
                    self.data[self.lb_sequences] = species
                    self.data[self.lb_A] = []
                    self.data[self.lb_B] = []
                    self.populate_list_with_items(species, self.lb_sequences, delete_all=True)
                    self.populate_list_with_items([], self.lb_A, delete_all=True)
                    self.populate_list_with_items([], self.lb_B, delete_all=True)
        return error, arg
    
    def open_case_file(self, case_file):
        error = fasta.OK
        if case_file:
            error, arg = self.case.load(case_file)
            if error == fasta.OK:
                self.fasta_file = self.case.data['filename']
                error, arg = self.S.load(self.fasta_file)
                if error == fasta.OK:
                    self.data[self.lb_sequences] = self.case.data["species"]
                    self.populate_list_with_items(self.case.data["species"],
                                                  self.lb_sequences, delete_all=True)
                    self.data[self.lb_A] = self.case.data["setA"]
                    self.populate_list_with_items(self.case.data["setA"],
                                                  self.lb_A, delete_all=True)
                    self.data[self.lb_B] = self.case.data["setB"]
                    self.populate_list_with_items(self.case.data["setB"],
                                                  self.lb_B, delete_all=True)
                    self.operation_method.set(self.case.data["operation"])
        return error, arg

    def cb_set_working_dir(self):
        self.cwd = filedialog.askdirectory(initialdir=self.cwd) or self.cwd
        self.setcwd(self.cwd)
        
    def cb_save_case_file(self):
        error = fasta.OK
        if self.case.data:
            case_file = filedialog.asksaveasfilename(defaultextension=".fc",
                                                     filetypes=[('case files', '.fc'), ('all files', '.*')],
                                                     initialdir=self.cwd,
                                                     #initialfile,
                                                     #message,
                                                     parent=self.root,
                                                     title="Save case file")
            if case_file:
                try:
                    self.case.save(case_file)
                except IOError:
                    error = fasta.ERROR_IO
                    arg = case_file
            else:
                return # cancel clicked. Ignore silently.    
        else:
            error = fasta.ERROR_NO_CASE_DATA
            arg = ''
        if error != fasta.OK:
            self.error_window(error, arg)
            
    def cb_about(self):
        self.about_window()
        
    def cb_help(self):
        self.help_window()
    
    def cb_save_report(self):
        error = fasta.OK
        out_file = filedialog.asksaveasfilename(defaultextension=".txt",
                                                filetypes=[('text files', '.txt'), ('all files', '.*')],
                                                initialdir=self.cwd,
                                                #initialfile,
                                                #message,
                                                parent=self.root,
                                                title="Save output")
        if not out_file:
            return # Cancel clicked, ignore silently
        try:
            with open(out_file,'w') as fp:
                fp.write(self.report.get(1., Tk.END))
        except FileNotFoundError:
            error = fasta.ERROR_FILE_NOT_FOUND
            arg = out_file
        except IOError:
            error = fasta.ERROR_IO
            arg = out_file
        except:
            error = fasta.ERROR_UNKNOWN
            arg = out_file
        if error != fasta.OK:
            self.error_window(error, arg)

    def cb_save_report_xls(self):
        error = fasta.OK
        out_file = filedialog.asksaveasfilename(defaultextension=".xls",
                                                filetypes=[('xls files', '.xls'), ('all files', '.*')],
                                                initialdir=self.cwd,
                                                #initialfile,
                                                #message,
                                                parent=self.root,
                                                title="Save output as xls")
        if not out_file:
            return # Cancel clicked, ignore silently
        self.reportxls.save(out_file)
        try:
            self.reportxls.save(out_file)
        except FileNotFoundError:
            error = fasta.ERROR_FILE_NOT_FOUND
            arg = out_file
        except IOError:
            error = fasta.ERROR_IO
            arg = out_file
        except:
            error = fasta.ERROR_UNKNOWN
            arg = out_file
        if error != fasta.OK:
            self.error_window(error, arg)

            
        
    def cb_clr(self):
        self.report.config(state=Tk.NORMAL)
        self.report.delete(1.0,Tk.END)
        self.report.config(state=Tk.DISABLED)
        self.reportxls.clear()
        
    def cb_run(self):
        operation = self.operation_method.get()
        set_A = self.S.select_sequences_from_list(self.data[self.lb_A])
        set_B = self.S.select_sequences_from_list(self.data[self.lb_B])
        if (operation <=2 and (len(set_A)==0 or len(set_B)==0)) or \
           (operation==3 and len(set_A)==0) or \
           (operation==4 and len(set_B)==0):
            return
        self.case.clear()
        try:
            self.case.populate(self.fasta_file,
                               self.data[self.lb_sequences],
                               self.data[self.lb_A],
                               self.data[self.lb_B],
                               operation)
        except AttributeError:
            return
        memofile = StringIO()
        report = fasta.Report(self.fasta_file, output_filename=memofile, reportxls = self.reportxls)
        
        if operation == 1:
            result = self.S.compare_sets(set_A, set_B)
            report.report_header(set_A, set_B)
            report.report_uniq_characters("Set A", set_A, set_B, result)
            report.report_footer()
            self.report.config(state=Tk.NORMAL)
            self.report.insert(Tk.END, memofile.getvalue())
            self.report.config(state=Tk.DISABLED)
        elif operation == 2:
            result = self.S.compare_sets(set_B, set_A)
            report.report_header(set_A, set_B)
            report.report_uniq_characters("Set B", set_B, set_A, result)
            report.report_footer()
            self.report.config(state=Tk.NORMAL)
            self.report.insert(Tk.END, memofile.getvalue())
            self.report.config(state=Tk.DISABLED)
        elif operation == 3:
            result = self.S.differences_within_set(set_A)
            report.report_header(set_A, set_B)
            report.report_differences_in_set("Set A", set_A, result)
            report.report_footer()
            self.report.config(state=Tk.NORMAL)
            self.report.insert(Tk.END, memofile.getvalue())
            self.report.config(state=Tk.DISABLED)
        elif operation == 4:
            result = self.S.differences_within_set(set_B)
            report.report_header(set_A, set_B)
            report.report_differences_in_set("Set B", set_B, result)
            report.report_footer()
            self.report.config(state=Tk.NORMAL)
            self.report.insert(Tk.END, memofile.getvalue())
            self.report.config(state=Tk.DISABLED)


def main():
    gui = Gui()
    gui.create_menu()
    gui.create_layout()
    gui.create_bindings()
    Tk.mainloop()
    gui.config.save()
    return 0






from collections import defaultdict
from itertools import zip_longest
import os
import re
import sys

import numpy as np
import xlwt

OK = 0b0000
ERROR_FILE_NOT_FOUND = 0b0001
ERROR_FILE_INVALID = 0b0010
ERROR_IO           = 0b0011
ERROR_NO_CASE_DATA = 0b0100
ERROR_UNEQUAL_SEQS = 0b0101
ERROR_CASE         = 0b0111
ERROR_UNKNOWN      = 0b1111


class Seq(object):
    ''' A class to hold the information of a single sequence '''
    S = "-?"+"".join([chr(i+65) for i in range(26)])
    ID = 0
    
    def __init__(self, hdr, data):
        self.ID, self.species = self.parse_hdr(hdr)
        self.data = data
        self.seq = self.parse_data(data)
        self.length = len(self.seq)

    def parse_data(self,data):
        ''' Parse the data string of the sequence '''
        seq=np.array([Seq.S.index(i) for i in data.strip()], dtype=int)
        return seq
    
    def parse_hdr(self, hdr):
        ''' Parse the header string of the sequence '''
        #>WBET042_Lyrodus_pedicellatus_Brittany_France
        if hdr[0] != ">":
            raise ValueError("Invalid header/file.")
        Seq.ID += 1
        species = hdr[1:].strip()
        return Seq.ID, species

    def __repr__(self):
        return "Seq object ({}, {})".format(self.ID, self.species)
    
    @staticmethod
    def pretty_print(m,c=" "):
        return c.join(Seq.S[_m] for _m in m)
        
        

class SequenceData(object):
    ''' Class to hold sequences '''
    def __init__(self):
        self.sequences = []

    def load(self, fn):
        ''' Load sequence data from filename *fn* '''
        sequences=[]
        error = OK
        arg = ''
        if not os.path.exists(fn):
            error = ERROR_FILE_NOT_FOUND
            arg = fn
            return error, arg
        # File can be opened...
        cnt = 0
        with open(fn) as f:
            while True:
                hdr = f.readline()
                if not hdr:
                    break
                cnt+=1
                try:
                    data = f.readline()
                except:
                    error = ERROR_IO
                    arg = hdr
                else:
                    cnt+=1
                    try:
                        sequences.append(Seq(hdr, data))
                    except ValueError:
                        error = ERROR_FILE_INVALID
                        arg = "Offending line: %d"%(cnt+1)
                        break
        if not error and not self.are_sequences_of_equal_lengths(sequences):
            error = ERROR_UNEQUAL_SEQS
        self.sequences = sequences
        return error, arg
                
        
    def are_sequences_of_equal_lengths(self, sequences):
        ''' Check whether all sequences are equally long.

        Raises ValueError if not all sequences are equally long.
        '''
        l = [s.length for s in sequences]
        return len(set(l))==1


    def get_species_info(self):
        ''' get_species_info()

        returns a dictionary with
            keys: species
            values: IDs
        '''
        d = defaultdict(lambda : [])
        for s in self.sequences:
            d[s.species].append(s.ID)
        return d

    def get_species_list(self):
        ''' get_species_list()

        returns list of sorted species names.
        '''
        d = self.get_species_info()
        k = list(d.keys())
        k.sort()
        return k


    def select_sequences_from_list(self, itemlist):
        group = [i for i in self.sequences if i.species in itemlist]
        return group
    
    def select_sequences(self, regex, invert = False, exclude=None):
        ''' select_sequences(regex, invert):

        regex: a regular expression or exact string
        invert: if True, the inverted selection is returned (not matching species)
        exclude: None or a regular expression to exclude these.

        returns a list with selected sequences.
        '''
        c = re.compile(regex)
        if not exclude:
            if invert:
                group = [i for i in self.sequences if not c.match(i.species)]
            else:
                group = [i for i in self.sequences if c.match(i.species)]
        else:
            x = re.compile(exclude)
            if invert:
                group = [i for i in self.sequences if not c.match(i.species) or x.match(i.species)]
            else:
                group = [i for i in self.sequences if c.match(i.species) and not x.match(i.species)]
            
        return group

    def select_two_sequence_sets(self, regex):
        ''' select_two_sequence_sets(regex)
        regex: a regular expression or exact string

        returns two sets of sequences: one matching set and one not matching set.
        The union fo the two sets is identical to the whole data set.
        '''
        set_A = self.select_sequences(regex, invert=False)
        set_B = self.select_sequences(regex, invert=True)
        return set_A, set_B
    
    def make_matrix(self, aset):
        # returns a matrix with sequence info.
        # ther returned shape seqeunce_length x number_of_seqs
        #
        return np.array([i.seq for i in aset]).T
    
    def comparison_within_set(self, aset, invert=False):
        ''' show_differences_within_set(aset)
        shows the columns within a set that are not identical, if any
        
        returns a list of tuples, of which each tuple is the column number, and the the column data.
        '''
        matrix = self.make_matrix(aset)
        return self.__compare_matrix(matrix, invert)
    
    def __compare_matrix(self, matrix, invert):
        sel = []
        for i, row in enumerate(matrix):
            if invert:
                if np.all(row==row[0]):
                    sel.append((i, row))
            else:
                if not np.all(row==row[0]):
                    sel.append((i, row))
        return sel

    def find_columns_matching_char_list(self, aset, char_list):
        n_list = [Seq.S.index(i) for i in char_list]
        m = self.make_matrix(aset)
        cols = [i for i,_m in enumerate(m) if any(any(_m==c) for c in n_list)]        
        return cols
        
    def differences_within_set(self, aset):
        ''' differences_within_set(aset)

        aset: list of seqeunces

        returns a list of those columns for which are one or more differences found between the sequences
        of this aset
        '''
        return self.comparison_within_set(aset, invert=False)

    def agreements_within_set(self, aset):
        ''' agreements_within_set(aset)

        aset: list of sequences

        returns a list of those columns for which all columns agree between the sequences 
        of this aset
        '''
        return self.comparison_within_set(aset, invert=True)
    
    def compare_sets(self, set_A, set_B, invert=False,
                       excluded_character_list_set_A='- ?'.split()):
        '''Compares set A with set B and list 
            the differences  if not invert
            the agreements if invert
        
        all columns that have coding letters that appear in
        excluded_characeter_list_set_A are excluded from comparison,
        as well ass all column for which set A is not consistent.
        '''
        
        # is set A composed of identical sequences?
        delta_A = self.differences_within_set(set_A)
        if delta_A:
            # columns that are different and should be excluded from comparison are
            cols_excl = [i[0] for i in delta_A]
        else:
            cols_excl = []
        cols_excl += self.find_columns_matching_char_list(set_A, excluded_character_list_set_A)
        matrix_A = self.make_matrix(set_A)[:,0]
        matrix_B = self.make_matrix(set_B)

        sel = []
        for i, (a, b) in enumerate(zip(matrix_A, matrix_B)):
            if i in cols_excl:
                continue
            if invert==False:
                if np.all(a!=b):
                    sel.append((i, a, b))
            else:
                if not np.all(a!=b):
                    sel.append((i, a, b))
        return sel


class Report(object):
    def __init__(self, filename=None, output_filename=None, reportxls = None):
        self.filename = filename or "noname"
        self.output_filename = output_filename or sys.stdout
        self.reportxls = reportxls
        try:
            self.reportxls.filename = self.filename
        except AttributeError:
            pass
            
    def report_unique_characters(self, set_A, set_B, differences_set_A, unique_characters_A):
        try:
            self.reportxls.report_unique_characters(set_A, set_B, differences_set_A, unique_characters_A)
        except AttributeError:
            pass
        w = self.output_filename
        w.write("Filename : {}\n".format(self.filename))
        w.write("-"*80+"\n")
        w.write("Set A:\n")
        for i,s in enumerate(set_A):
            w.write("%2d %d %s\n"%(i+1, s.ID, s.species))
        w.write("\n")
        w.write("Set B:\n")
        for i,s in enumerate(set_B):
            w.write("%2d %d %s\n"%(i+1, s.ID, s.species))
        w.write("-"*80+"\n")
        w.write("\n\n")
        if differences_set_A:
            w.write("The differences WITHIN the set A are:\n\n")
            w.write("column: chars\n")
            for _d in differences_set_A:
                w.write("%6d %s\n"%(_d[0]+1, Seq.pretty_print(_d[1])))
            w.write("\n")
        if unique_characters_A:
            w.write("Unique characters of set A:\n\n")
            w.write("column: chr|  characters different in other species\n")
            w.write("           |  ")
            for i in range(len(set_B)):
                w.write("%d "%((i+1)%10))
            w.write("\n")
            w.write("-"*80+"\n")
            for _d in unique_characters_A:
                w.write("%6d: %s  |  "%(_d[0]+1, Seq.pretty_print([_d[1]])))
                w.write("%s\n"%(Seq.pretty_print(_d[2])))
            w.write("-"*80+"\n")
            a = len(unique_characters_A)
            b = len(set_A)
            f = a/b*100
            w.write("%d of %d characters are unique (%.1f%%)"%(a,b,f))
        else:
            w.write("Set A has no unique characters\n")
        if not w == sys.stdout:
            w.close()
        else:
            w.write("\n")

    def report_header(self, set_A, set_B):
        try:
            self.reportxls.report_header(set_A, set_B)
        except AttributeError:
            pass
        w = self.output_filename
        w.write("Filename : {}\n".format(self.filename))
        w.write("-"*80+"\n")
        w.write("Set A:\n")
        for i,s in enumerate(set_A):
            w.write("%2d %d %s\n"%(i+1, s.ID, s.species))
        w.write("\n")
        w.write("Set B:\n")
        for i,s in enumerate(set_B):
            w.write("%2d %d %s\n"%(i+1, s.ID, s.species))
        w.write("-"*80+"\n")
        w.write("\n\n")

    def report_footer(self):
        try:
            self.reportxls.report_footer()
        except AttributeError:
            pass
        w = self.output_filename
        w.write("\n")
        w.write("="*80)
        w.write("\n\n")
        
    def report_uniq_characters(self, set_name, set_A, set_B, unique_characters_A):
        try:
            self.reportxls.report_uniq_characters(set_name, set_A, set_B, unique_characters_A)
        except AttributeError:
            pass
        w = self.output_filename
        if unique_characters_A:
            w.write("Characters unique in set {}, but different in other species set:\n\n".format(set_name))
            w.write("column: chr|  characters different in other species\n")
            w.write("           |  ")
            for i in range(len(set_B)):
                w.write("%d "%((i+1)%10))
            w.write("\n")
            w.write("-"*80+"\n")
            for _d in unique_characters_A:
                w.write("%6d: %s  |  "%(_d[0]+1, Seq.pretty_print([_d[1]])))
                w.write("%s\n"%(Seq.pretty_print(_d[2])))
            a = len(unique_characters_A)
            b = len(set_A[0].data)
            f = a/b*100
            w.write("\n%d of %d characters are unique (%.1f%%)"%(a,b,f))
        else:
            w.write("{} has no unique characters\n".format(set_name))

    def report_differences_in_set(self, set_name, set_A, differences_set):
        try:
            self.reportxls.report_differences_in_set(set_name, differences_set)
        except IOError:
            pass
        w = self.output_filename
        if differences_set:
            w.write("The differences within {} are:\n\n".format(set_name))
            w.write("column: chars\n")
            w.write("-"*80)
            w.write("\n")
            for _d in differences_set:
                w.write("%6d %s\n"%(_d[0]+1, Seq.pretty_print(_d[1])))
            w.write("\n")
            a = len(differences_set)
            b = len(set_A[0].data)
            f = (b-a)/b*100
            w.write("\n%d of %d characters are different (%.1f%% identical)"%(a,b,f))

        else:
            w.write("All sequences within {} are identical.\n".format(set_name))
            

class MyWorkbook(xlwt.Workbook):
    def __init__(self, *p, **kw):
        super().__init__(*p, *kw)

    def remove_all_sheets(self):
        self._Workbook__worksheet_idx_from_name.clear()
        self._Workbook__worksheets.clear()
        
            
class ReportXLS(object):
    def __init__(self):
        self.book = MyWorkbook()
        self.sheet_idx = 0
        self.__row = 0
        
    def __create_sheet(self):
        sheet_name = "worksheet%02d"%(self.sheet_idx)
        self.sheet = self.book.add_sheet(sheet_name)
        self.sheet_idx+=1
        
    def clear(self):
        self.book.remove_all_sheets()
        self.sheet_idx = 0
        
    def save(self, fn):
        try:
            self.book.save(fn)
        except IndexError:
            pass # try to save an empty workbook ignore.
        
    def report_header(self, set_A, set_B):
        self.__create_sheet()
        n = 0
        self.sheet.write(n, 0, "Filename:")
        self.sheet.write(n, 1, self.filename)
        n+=2
        self.sheet.write(n, 0, "Set A")
        self.sheet.write(n, 1, "Set B")
        n+=1
        for a, b in zip_longest(set_A, set_B):
            if a:
                self.sheet.write(n, 0, "%s (%d)"%(a.species, a.ID))
            if b:
                self.sheet.write(n, 1 ,"%s (%d)"%(b.species, b.ID))
            n+=1
        self.__row = n

    def report_footer(self):
        pass
    
    def report_uniq_characters(self, set_name, set_A, set_B, unique_characters_A):
        n = self.__row +2
        self.sheet.write(n, 0, "Unique characters of {}:\n\n".format(set_name))
        n+=1
        self.sheet.write(n, 1, "Column")
        self.sheet.write(n, 2, "Character")
        self.sheet.write(n, 3, "Characters in different in other species")
        n+=1
        for i, v in enumerate(set_B):
            self.sheet.write(n, 3+i, "%d"%(i+1))
        n+=1
        for _d in unique_characters_A:
            self.sheet.write(n, 1, "%d"%(_d[0]+1))
            self.sheet.write(n, 2, Seq.pretty_print([_d[1]]))
            for i, _dd in enumerate(_d[2]):
                self.sheet.write(n, 3+i, Seq.pretty_print([_dd]))
            n+=1
        self.__row = n
        
    def report_differences_in_set(self, set_name, differences_set):
        n = self.__row +2
        self.sheet.write(n, 0, "The differences within {} are:".format(set_name))
        n+=1
        self.sheet.write(n, 1, "Column")
        self.sheet.write(n, 2, "Chars")
        n+=1
        for _d in differences_set:
            self.sheet.write(n, 1, "%d"%(_d[0]+1))
            for i, _dd in enumerate(_d[1]):
                self.sheet.write(n, 2+i, Seq.pretty_print([_dd]))
            n+=1





ERRORS = {0b0001:'File not found.',
          0b0010:'Invalid or corrupt file.',
          0b0011:'Could not save file.',
          0b0100:'No case data.',
          0b0101:'Not all sequences are of equal length',
          0b0111:'Case file formatting error',
          0b1111:'Unkwown error.',
}



ABOUT_TEXT="""
ABOUT

FastaChar is a simple python program that reads a fasta file
with sequences for a number of different species. From the list
of species, the user can make a selection of two sets (A) and 
(B), which can be analysed. 

The available operations are
* Unique characters A: 
  lists the columns in the sequences of A for which holds that:
      - they are the same for all sequences in set A
      - all sequences in set B have a different nucleotide or amino 
        acid in this column.
* Differences within A:
  lists the columns of all sequences where the nucleotide or amino
  acid is not identical.

This software is licensed under GPLv3.

Authors:
    Lucas Merckelbach lucas.merckelbach@hzg.de
    Luisa Borges      luisaborges2000@yahoo.co.uk

April 2017
"""

DISCLAIMER="""
DISCLAIMER

This software is provided "as is" and any expressed or implied
warranties, including, but not limited to, the implied warranties of
merchantability and fitness for a particular purpose are
disclaimed. In no event shall the regents or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

"""

HELP_TEXT="""
HELP

FastaChar is a simple python program to compare aligned sequences of
different or (supposedly) the same species.

The sequences are assumed to be provided in the standard fasta file
format. See for example https://en.wikipedia.org/wiki/FASTA_format.

WORKFLOW
========

Before using the program the user needs to prepare a valid fasta file,
with sequences of one or more species. A given species may be 
represented by more than one sequence, each with its own ID.

To open a fasta file, select from the menu "File" and then "Open fasta
file". A dialogue window appears from which a fasta file can be
selected. The directory displayed is the working directory (home
directory if not previously specified). Prior to opening the fasta
file, it may be convenient to set the working directory path
first. The setting of the working directory is persistent (i.e., it
will be kept from one session to the next).


When the fasta file is opened, all species distinguished in the fasta
file will appear in the top left list box (species).

NOTES: 

   1) if a species has multiple sequences, it will still appear as one
      entry.  

   2) if a given species has multiple sequences, and the names are not
      exactly the same, then they will appear as two or more different 
      species.

From the pool of available species, the user can make two selections,
called set A and set B. Both sets should contain at least one
species. When species are selected, they are *moved* from their origin
(the box called "species" in this case) to their destination (set A for
example).

See below (MAKING A SELECTION) how to
make a selection of one or more species and drag them into a list box.

OPERATION
=========
The next step is to choose an operation. Two types of operations are
available:
    1) list unique characters of a set
    2) list the differences within a set.

Set A can be compared with set B, or vice versa, hence giving rise to
four operations in total.

Unique characters A operation 
-----------------------------
This operation compares the sequences
in set A with those in set B and lists all columns and base
characters (nucleotides or aminoacids) for which holds that for a given
column:
    - all base characters in set A are identical; and
    - all base characeters in set B are different from those in set A

Differences within A operation 
------------------------------
This operation lists all those columns for which the base characters
(nucleotides or amino acids) of the sequences in set A are *not* the same.

After choosing the operation, the "PROCESS" button can be hit, to make
the comparsion. The result report appears in the bottom window.

RESULT WINDOW
=============

The results are displayed in the following format. First the selected
fasta file is printed, followed by all the sequences (including ID) in
set A and set B. Then the actual result is printed. 

If set A is found not to have any unique characters then that will
be the result. Otherwise the result is printed as


Unique characters of set A:

column: chr|  characters different in other species
           |  1 2 3 4 5 6 7 8 
--------------------------------------------------------------------------------
    32: G  |  T T T T T T T T
   202: G  |  C C C C C C C C
   207: G  |  C C C C C C C C


which means that columns 32, 202 and 207 in *all* sequences of set A
are the same, and have the character G. Set B, having 8 sequences,
has sequences that are all different from G. Using the list of
sequences in set B printed above, the differences between the
sequences in this set can be attributed to different sequences.

If the operation chosen was "differences within a set", then result is similar.

The differences within set A are:

column: chars
    17 T T A A
    26 A A C C
   184 G G T T
   189 C C T T
   195 T T G G
   206 C C A A
   216 A A C C
   234 T T C C

which indicates the columns where the sequences in set A (in this case) differ.

Repeated processing will add the result to the result window, which
can be cleared using the "Clear output" button. The information in the
result window can be written to a text file via menu Output -> Save
report.

CASE FILES
==================

A given setting, defined by the selections made for species, set A,
set B and operation, can be saved to file for reanalysis later, or
as starting point for a similar, but not identical comparison. A case
file can be saved using the menu File -> Save case file, or loaded via
menu File -> Load case file.


MAKING A SELECTION 
==================

To make a selection, click with the mouse on a species name. To select
multiple species, click on another species name whilst pressing
"SHIFT"; this will select all the species in between as well. To add
more species to the selection, click on the species to select, whilst
pressing "CTRL". Use the right-hand button to drag the selection into
the listbox of choice.

This works identical for making a selection to move species from the
list "species" to set A, or back, or between set A and B.

"""



# Example script how to do an analysis of a fasta file accessing the
# fasta module directly, and not using a graphical interface. The
# example script reads a fasta file, and divides all the species in a
# two sets, one that with species names that match a regular
# expression, and a set with sequences that does not match the regular
# expression. Then, for set A, the differences within this set as well
# as its unique characters are computed. Finally, the results are
# reported and dumped on the terminal.

from fasta import fasta


filename = "../data/COI_sequences_MUSCLE.fas"

S = fasta.SequenceData()

S.load(filename)

species = S.get_species_list()
print("Species in this file:")
for s in species:
    print("{:30s}".format(s))
print()

# Divide all the species in two groups, set A that matches the regex,
# and set B that does not.
set_A, set_B = S.select_two_sequence_sets("Lyrodus.pedicellatus.*[mM][Ss]")

# Compute the differences within set A
differences_set_A = S.differences_within_set(set_A)

# Compute the unique characters in A with respect to B
unique_characters_A = S.compare_sets(set_A, set_B)

# Report the results to the terminal.
report = fasta.Report(filename)
report.report_unique_characters(set_A, set_B, differences_set_A, unique_characters_A)



from setuptools import setup

setup(name="fastachar",
      version="0.0.1",
      packages = ['fasta'],
      py_modules = ['fastachar'],
      entry_points = {'console_scripts':[],#['fastachar_gui = fastachar_gui:main'],
                      'gui_scripts':['fastachar = fastachar:main']
                      },
      install_requires = 'numpy xlwt'.split(),
      author="Lucas Merckelbach",
      author_email="lucas.merckelbach@hzg.de",
      description="A simple program with GUI to compare dna sequences",
      long_description="""A simple python module with graphical user interface
that allows to compare a group of aligned DNA sequences with another group of
aligned DNA sequences. Different operations can be specified such as find
differences or correspondences.""",
    url='http://cubic-l.science/fastachar.html',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: MacOS :: MacOS X',
                  'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
    ])
